from io import BytesIO
from unittest.mock import call, patch, sentinel
from uuid import UUID

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from frgal.utils import make_awaitable
from regal import NssetId
from regal.exceptions import ObjectDoesNotExist

from ferda.constants import LogEntryType, LoggerResult, LogReferenceType
from ferda.tests.utils import LoggerMixin

object_uuid = str(UUID(int=1))


@override_settings(ROOT_URLCONF='ferda.tests.urls', FERDA_RECORD_STATEMENT_TIMEOUT=sentinel.timeout)
class NssetRecordStatementTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = get_user_model().objects.create_superuser('admin')  # type: ignore
        self.client.force_login(user)

    @patch('ferda.registry.nsset.api.STATEMENTOR')
    def test_get_nsset_pdf(self, nsset_statement_mock):
        nsset_statement_mock.get_nsset_statement.return_value = make_awaitable(BytesIO(b'PDF content'))
        response = self.client.get(
            reverse('api-1:get_nsset_record_statement_pdf', kwargs={'nsset_id': object_uuid})
        )
        self.assertLog(
            LogEntryType.RECORD_STATEMENT,
            LoggerResult.SUCCESS,
            input_references={LogReferenceType.NSSET: object_uuid},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['Content-Disposition'], 'attachment; filename="record-statement.pdf"')
        self.assertEqual(response.content, b'PDF content')
        self.assertEqual(nsset_statement_mock.mock_calls, [
            call.get_nsset_statement(NssetId(object_uuid), private=True, timeout=sentinel.timeout),
        ])

    @patch('ferda.registry.nsset.api.STATEMENTOR')
    def test_nsset_not_found(self, nsset_statement_mock):
        nsset_statement_mock.get_nsset_statement.side_effect = ObjectDoesNotExist
        response = self.client.get(
            reverse('api-1:get_nsset_record_statement_pdf', kwargs={'nsset_id': object_uuid})
        )
        self.assertLog(
            LogEntryType.RECORD_STATEMENT,
            LoggerResult.FAIL,
            input_references={LogReferenceType.NSSET: object_uuid},
        )
        self.assertEqual(response.status_code, 404)
        self.assertEqual(nsset_statement_mock.mock_calls, [
            call.get_nsset_statement(NssetId(object_uuid), private=True, timeout=sentinel.timeout),
        ])
